################################################################################
# Copyright (C) 2020-2021 Abstract Horizon
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the Apache License v2.0
# which accompanies this distribution, and is available at
# https://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#    Daniel Sendula - initial API and implementation
#
#################################################################################
from typing import Tuple, Callable, Union, Optional

from pygame import Rect
from pygame.font import Font

from pyros_support_ui.components import Component, Collection, ALIGNMENT, UiHint, LeftRightLayout, BaseUIFactory

Number = Union[int, float]
GetterSetterPair = Tuple[Callable[[], Number], Callable[[Number], None]]


class NumberInputComponent(Collection):

    def __init__(self,
                 rect: Rect,
                 ui_factory: BaseUIFactory,
                 getter_setter_pair: GetterSetterPair,
                 name: str,
                 minimal_value: Optional[Number] = None, maximum_value: Optional[Number] = None,
                 button_font: Optional[Font] = None, value_font: Optional[Font] = None,
                 top_scale: int = 7, bottom_scale: int = 4):

        super(NumberInputComponent, self).__init__(rect, layout=LeftRightLayout(margin=5))
        self.getter = getter_setter_pair[0]
        self.setter = getter_setter_pair[1]
        self.name = name
        self.minimal_value = minimal_value
        self.maximum_value = maximum_value
        self.button_font = button_font if button_font is not None else ui_factory.font
        self.value_font = value_font if value_font is not None else ui_factory.font
        self.height = 40 if rect is None or rect.height == 0 else rect.height

        self.widths = [self.button_font.size(w)[0] for w in ["<", "<<", ">>", ">"]]
        self.margin = 3

        self.top_scale = top_scale
        self.bottom_scale = bottom_scale
        self.scale = [1]
        p = 0.1
        for _ in range(1, bottom_scale):
            self.scale.insert(0, p)
            p /= 10.0

        p = 10.0
        for _ in range(1, top_scale):
            self.scale.append(p)
            p *= 10.0

        self.add_component(ui_factory.text_button(Rect(rect.x, rect.y, self.widths[0], self.height), "<", self.on_click_minus1, font=self.button_font, hint=UiHint.NO_DECORATION))
        self.add_component(ui_factory.text_button(Rect(rect.x + 34, rect.y, self.widths[1], self.height), "<<", self.on_click_minus01, font=self.button_font, hint=UiHint.NO_DECORATION))
        self.add_component(Component(Rect(0, 0, 5, self.height)))
        self.add_component(ui_factory.label(Rect(rect.x + 112, rect.y, 30, self.height), self.name, v_alignment=ALIGNMENT.MIDDLE, font=self.value_font))

        self.left = ui_factory.label(Rect(rect.x + 100, rect.y, 64, self.height), "", h_alignment=ALIGNMENT.RIGHT, v_alignment=ALIGNMENT.MIDDLE)
        self.right = ui_factory.label(Rect(rect.x + 164, rect.y, 30, self.height), "", h_alignment=ALIGNMENT.LEFT, v_alignment=ALIGNMENT.MIDDLE)
        self.add_component(self.left)
        self.add_component(self.right)

        self.add_component(ui_factory.text_button(Rect(rect.right - 30 - 34, rect.y, self.widths[2], self.height), ">>", self.on_click_plus01, font=self.button_font, hint=UiHint.NO_DECORATION))
        self.add_component(ui_factory.text_button(Rect(rect.right - 30, rect.y, self.widths[3], self.height), ">", self.on_click_plus1, font=self.button_font, hint=UiHint.NO_DECORATION))

    def redefine_rect(self, rect):
        super(NumberInputComponent, self).redefine_rect(rect)
        # self.rect = rect
        #
        # self.components[3].redefine_rect(Rect(rect.x, rect.y, self.widths[0], self.height))
        # self.components[4].redefine_rect(Rect(self.components[3].rect.right + self.margin, rect.y, self.widths[1], self.height))
        # self.components[5].redefine_rect(Rect(self.components[4].rect.right + self.margin, rect.y, self.widths[2], self.height))
        #
        # self.components[1].redefine_rect(Rect(self.components[5].rect.right + self.margin, rect.y, 64, self.height))
        # self.components[0].redefine_rect(Rect(rect.x + 112, rect.y, 30, self.height))
        # self.components[2].redefine_rect(Rect(rect.x + 164, rect.y, 30, self.height))
        #
        # self.components[6].redefine_rect(Rect(rect.right - 30 - 34 * 2, rect.y, self.widths[3], self.height))
        # self.components[7].redefine_rect(Rect(rect.right - 30 - 34, rect.y, self.widths[4], self.height))
        # self.components[8].redefine_rect(Rect(rect.right - 30, rect.y, self.widths[5], self.height))

    def draw(self, surface):
        value = self.getter()

        if value is not None:
            try:
                if abs(value > 0.1):
                    s = ("   {0:." + str(self.bottom_scale) + "f}").format(value)
                    # s = "{0:.2f}".format(value)
                    i = s.find('.')
                    i = i if i >= 0 else len(s)
                    left = s[0:i + 1]
                    right = s[i + 1:]
                    self.left.text = left
                    self.right.text = right
                    self.left.h_alignment = ALIGNMENT.RIGHT
                else:
                    self.left.text = ("   {0:." + str(self.bottom_scale) + "f}").format(value)
                    self.right.text = ""
                    self.left.h_alignment = ALIGNMENT.LEFT
            except TypeError:
                pass
                # print(f"Cannot handle value {value} it is not a number ({type(value)}")
        else:
            self.left.text = "-"
            self.right.text = ""

        super().draw(surface)

    def set_new_value(self, value) -> bool:
        if self.minimal_value is not None and value < self.minimal_value:
            return False
        if self.maximum_value is not None and value > self.maximum_value:
            return False

        self.setter(value)
        return True

    def on_click_plus1(self, *_args):
        value = self.getter()
        for i in range(1, len(self.scale)):
            if abs(value) < self.scale[i]:
                value += self.scale[i - 1]
                self.set_new_value(value)
                return

        value += self.scale[-1]
        self.set_new_value(value)

    def on_click_minus1(self, *_args):
        value = self.getter()
        for i in range(1, len(self.scale)):
            if abs(value) < self.scale[i]:
                value -= self.scale[i - 1]
                self.set_new_value(value)
                return

        value -= self.scale[-1]
        self.set_new_value(value)

    def on_click_plus01(self, *_args):
        value = self.getter()
        for i in range(2, len(self.scale)):
            if abs(value) < self.scale[i]:
                value += self.scale[i - 2]
                self.set_new_value(value)
                return
        value += self.scale[-2]
        self.set_new_value(value)

    def on_click_minus01(self, *_args):
        value = self.getter()
        for i in range(2, len(self.scale)):
            if abs(value) < self.scale[i]:
                value -= self.scale[i - 2]
                self.set_new_value(value)
                return
        value -= self.scale[-2]
        self.set_new_value(value)



################################################################################
# Copyright (C) 2020 Abstract Horizon
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the Apache License v2.0
# which accompanies this distribution, and is available at
# https://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#    Daniel Sendula - initial API and implementation
#
#################################################################################

import pygame
import os
import time

BLACK = (0, 0, 0)
DARK_GREY = (80, 80, 80)
GREY = (160, 160, 160)
WHITE = (255, 255, 255)
LIGHT_BLUE = (160, 255, 255)
GREEN = (128, 255, 128)
RED = (255, 128, 128)
YELLOW = (255, 255, 128)
BLUE = (128, 128, 255)

smallFont = None
font = None
big_font = None

frameclock = pygame.time.Clock()
screen = None
backgroundImage = None
scaledBackground = None
gccFullImage = None
gccGreenImage = None
gccColouredImage = None

SCREEN_RECT = None


def _this_path(filename):
    return os.path.join(os.path.dirname(__file__), filename)


def init_all(screen_size, load_background=False, title="GCC"):
    global screen, backgroundImage, scaledBackground
    global smallFont, font, big_font
    global SCREEN_RECT
    global gccFullImage, gccGreenImage, gccColouredImage

    pygame.init()

    font_file = _this_path("garuda.ttf")
    big_font = pygame.font.Font(font_file, 20)
    font = pygame.font.Font(font_file, 16)
    smallFont = pygame.font.Font(font_file, 12)

    screen = pygame.display.set_mode(screen_size)
    pygame.display.set_caption(title)

    SCREEN_RECT = pygame.Rect(0, 0, screen.get_width(), screen.get_height())

    if load_background:
        backgroundImage = pygame.image.load(_this_path("blue-background.png"))
        scaledBackground = pygame.transform.scale(backgroundImage, screen.get_size())

    gccFullImage = pygame.image.load(_this_path("GCC_full.png"))
    gccGreenImage = pygame.image.load(_this_path("GCC_green_small.png"))
    gccColouredImage = pygame.image.load(_this_path("GCC_coloured_small.png"))

    return screen


def screen_resized(size):
    global scaledBackground

    scaledBackground = pygame.transform.scale(backgroundImage, size)


def background(top_frame=False):
    if scaledBackground is not None:
        screen.blit(scaledBackground, (0, 0))
    else:
        screen.fill((0, 0, 0))

    if top_frame:
        draw_top_frame()


def frame_end():
    pygame.display.flip()
    frameclock.tick(30)


def draw_key_value(key, value, x, y, colour=WHITE):
    screen.blit(font.render(key + ":", 1, colour), (x, y))
    screen.blit(font.render(value, 1, colour), (x + 100, y))
    return y + 20  # font.get_height()


def draw_big_text(text, pos, colour=WHITE):
    screen.blit(big_font.render(text, 1, colour), pos)


def draw_text(text, pos, colour=WHITE):
    screen.blit(font.render(text, 1, colour), pos)


def draw_small_text(text, pos, colour=WHITE):
    screen.blit(smallFont.render(text, 1, colour), pos)


def draw_rect(pos, size, colour=LIGHT_BLUE, stick=0, outside=1):

    pygame.draw.line(screen, colour, (pos[0] - stick, pos[1] - outside), (pos[0] + size[0] + stick, pos[1] - outside))
    pygame.draw.line(screen, colour, (pos[0] - stick, pos[1] + size[1] + outside), (pos[0] + size[0] + stick, pos[1] + size[1] + outside))

    pygame.draw.line(screen, colour, (pos[0] - outside, pos[1] - stick), (pos[0] - outside, pos[1] + size[1] + stick))
    pygame.draw.line(screen, colour, (pos[0] + size[0] + outside, pos[1] - stick), (pos[0] + size[0] + outside, pos[1] + size[1] + stick))


def draw_frame(rect, colour=LIGHT_BLUE, background_colour=BLACK):
    if background_colour is not None:
        pygame.draw.rect(screen, background_colour, rect, 0)
    pygame.draw.rect(screen, colour, (rect[0] + 4, rect[1] + 4, rect[2] - 8, rect[3] - 8), 1)
    for i in range(6, 12):
        pygame.draw.line(screen, colour, (rect[0] + rect[2] - i, rect[1] + rect[3] - 6), (rect[0] + rect[2] - 6, rect[1] + rect[3] - i), 1)


def draw_top_frame(colour=LIGHT_BLUE):
    screen_rect = SCREEN_RECT

    if int(time.time()) % 20 < 10:
        screen.blit(gccGreenImage, (screen_rect[0] + screen_rect[2] - 8 - gccGreenImage.get_width(), screen_rect[1] + 8))
    else:
        screen.blit(gccColouredImage, (screen_rect[0] + screen_rect[2] - 8 - gccGreenImage.get_width(), screen_rect[1] + 8))

    pygame.draw.line(screen, colour, (screen_rect[0] + 4, screen_rect[1] + 4), (screen_rect[0] + 330, screen_rect[1] + 4))
    pygame.draw.line(screen, colour, (screen_rect[0] + 4, screen_rect[1] + 4), (screen_rect[0] + 4, screen_rect[1] + 30))
    pygame.draw.line(screen, colour, (screen_rect[0] + 4, screen_rect[1] + 30), (screen_rect[0] + 300, screen_rect[1] + 30))
    pygame.draw.line(screen, colour, (screen_rect[0] + 300, screen_rect[1] + 30), (screen_rect[0] + 330, screen_rect[1] + 4))

    pygame.draw.line(screen, colour, (screen_rect[0] + 330, screen_rect[1] + 4), (screen_rect[0] + screen_rect[2] - 12, screen_rect[1] + 4))

    pygame.draw.line(screen, colour, (screen_rect[0] + screen_rect[2] - 12, screen_rect[1] + 4), (screen_rect[0] + screen_rect[2] - 4, screen_rect[1] + 12))

    pygame.draw.line(screen, colour, (screen_rect[0] + screen_rect[2] - 4, screen_rect[1] + 12), (screen_rect[0] + screen_rect[2] - 4, screen_rect[1] + screen_rect[3] - 4))
    pygame.draw.line(screen, colour, (screen_rect[0] + 4, screen_rect[1] + screen_rect[3] - 4), (screen_rect[0] + screen_rect[2] - 4, screen_rect[1] + screen_rect[3] - 4))
    pygame.draw.line(screen, colour, (screen_rect[0] + 4, screen_rect[1] + screen_rect[3] - 4), (screen_rect[0] + 4, screen_rect[1] + 30))

    pygame.draw.line(screen, colour, (screen_rect[0] + screen_rect[2] - gccGreenImage.get_width() - 4, screen_rect[1] + gccGreenImage.get_height() + 12),
                     (screen_rect[0] + screen_rect[2] - 4, screen_rect[1] + gccGreenImage.get_height() + 12))
    pygame.draw.line(screen, colour, (screen_rect[0] + screen_rect[2] - gccGreenImage.get_width() - 4, screen_rect[1] + gccGreenImage.get_height() + 12),
                     (screen_rect[0] + screen_rect[2] - gccGreenImage.get_width() - 12, screen_rect[1] + gccGreenImage.get_height() + 4))

    pygame.draw.line(screen, colour, (screen_rect[0] + screen_rect[2] - gccGreenImage.get_width() - 12, screen_rect[1] + 4),
                     (screen_rect[0] + screen_rect[2] - gccGreenImage.get_width() - 12, screen_rect[1] + gccGreenImage.get_height() + 4))


def draw_filled_rect(pos, size, colour=LIGHT_BLUE):
    pygame.draw.rect(screen, colour, pygame.Rect(pos, size))


def draw_image(image, pos, stick=6):
    screen.blit(image, pos)

    draw_rect(pos, image.get_size(), stick=stick, outside=1)


def draw_up_arrow(x1, y1, x2, y2, colour):
    pygame.draw.line(screen, colour, (x1, y2), (x1 + (x2 - x1) / 2, y1))
    pygame.draw.line(screen, colour, (x1 + (x2 - x1) / 2, y1), (x2, y2))
    pygame.draw.line(screen, colour, (x1, y2), (x2, y2))


def draw_down_arrow(x1, y1, x2, y2, colour):
    pygame.draw.line(screen, colour, (x1, y1), (x1 + (x2 - x1) / 2, y2))
    pygame.draw.line(screen, colour, (x1 + (x2 - x1) / 2, y2), (x2, y1))
    pygame.draw.line(screen, colour, (x1, y1), (x2, y1))


def draw_graph(pos, size, data, min_data, max_data, max_points, axis_colour=LIGHT_BLUE, data_colour=GREEN, stick=0):

    def calc_point(_d, mn, mx, _x, _y, h):
        r = mx - mn - 1
        if r < 5:
            r = 5

        _d = _d - mn
        if _d > mx:
            _d = mx
        if _d < 0:
            _d = 0
        _d = r - _d

        return _x, _y + int(_d / r * h)

    outside = 0

    # pygame.draw.line(screen, axisColour, (pos[0] - stick, pos[1] - outside), (pos[0] + size[0] + stick, pos[1] - outside))
    pygame.draw.line(screen, axis_colour, (pos[0] - stick, pos[1] + size[1] + outside), (pos[0] + size[0] + stick, pos[1] + size[1] + outside))

    pygame.draw.line(screen, axis_colour, (pos[0] - outside, pos[1] - stick), (pos[0] - outside, pos[1] + size[1] + stick))
    # pygame.draw.line(screen, axisColour, (pos[0] + size[0] + outside, pos[1] - stick), (pos[0] + size[0] + outside, pos[1] + size[1] + stick))

    if len(data) > 1:
        to = len(data) - max_points
        if to < 0:
            to = 0

        stretch = size[0] / min([max_points, len(data) - 1])

        x = pos[0] + size[0] - 1

        _ld = data[len(data) - 1]
        lp = calc_point(_ld, min_data, max_data, x, pos[1], size[1])

        for i in range(len(data) - 2, to - 1, -1):
            x = x - int(stretch)
            d = data[i]
            p = calc_point(d, min_data, max_data, x, pos[1], size[1])

            pygame.draw.line(screen, data_colour, lp, p)
            _ld = d
            lp = p



################################################################################
# Copyright (C) 2020 Abstract Horizon
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the Apache License v2.0
# which accompanies this distribution, and is available at
# https://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#    Daniel Sendula - initial API and implementation
#
#################################################################################
import pygame.color

from pyros_support_ui.components import *


class ButtonDecoration(Component):
    def __init__(self, colour):
        super(ButtonDecoration, self).__init__(None)  # Call super constructor to store rectable
        self.colour = colour

    def draw(self, surface):
        pygame.draw.rect(surface, self.colour, self.rect)


class BorderDecoration(Component):
    def __init__(self, colour):
        super(BorderDecoration, self).__init__(None)  # Call super constructor to store rectable
        self.colour = colour

    def draw(self, surface):
        pygame.draw.rect(surface, self.colour, self.rect, 1)


class FlatThemeFactory(BaseUIFactory):
    def __init__(self, ui_adapter, font=None, small_font=None,
                 colour=pygame.color.THECOLORS['cyan'],
                 warning_colour=pygame.color.THECOLORS['orange'],
                 error_colour=pygame.color.THECOLORS['red'],
                 disabled_colour=pygame.color.THECOLORS['darkblue'],
                 background_colour=pygame.color.THECOLORS['gray32'],
                 mouse_over_colour=pygame.color.THECOLORS['lightgray'],
                 mouse_over_background_colour=pygame.color.THECOLORS['gray32'],
                 warning_mouse_over_background_colour=pygame.color.THECOLORS['darkorange4'],
                 error_mouse_over_background_colour=pygame.color.THECOLORS['darkred']
                 ):
        super(FlatThemeFactory, self).__init__(ui_adapter,
                                               font=font,
                                               small_font=small_font,
                                               colour=colour,
                                               warning_colour=warning_colour,
                                               error_colour=error_colour,
                                               disabled_colour=disabled_colour,
                                               background_colour=background_colour,
                                               mouse_over_colour=mouse_over_colour,
                                               mouse_over_background_colour=mouse_over_background_colour,
                                               warning_mouse_over_background_colour=warning_mouse_over_background_colour,
                                               error_mouse_over_background_colour=error_mouse_over_background_colour)

    def setMouseOverColour(self, mouse_over_colour):
        self.mouse_over_colour = mouse_over_colour

    def setBackgroundColour(self, background_colour):
        self.background_colour = background_colour

    def label(self, rect, text, font=None, colour=None, h_alignment=ALIGNMENT.LEFT, v_alignment=ALIGNMENT.TOP, hint=UiHint.NORMAL):
        return Label(rect, text, font=font if font is not None else self.font, colour=colour, h_alignment=h_alignment, v_alignment=v_alignment)

    def _disabled_label(self, rect, text, font=None, h_alignment=ALIGNMENT.LEFT, v_alignment=ALIGNMENT.TOP, hint=UiHint.NORMAL):
        return Label(rect, text, font=font if font is not None else self.font, colour=self.disabled_colour, h_alignment=h_alignment, v_alignment=v_alignment)

    def image(self, rect, image, h_alignment=ALIGNMENT.LEFT, v_alignment=ALIGNMENT.TOP, hint=UiHint.NORMAL):
        return Image(rect, image, h_alignment=h_alignment, v_alignment=v_alignment)

    def button(self, rect, on_click=None, on_hover=None, label=None, disabled_label=None, hint=UiHint.NORMAL):
        background_colour = self.background_colour
        mouse_over_colour = self.mouse_over_colour
        if hint == UiHint.WARNING:
            background_colour = pygame.color.THECOLORS['darkorange3']
            mouse_over_colour = pygame.color.THECOLORS['darkorange']
        elif hint == UiHint.ERROR:
            background_colour = pygame.color.THECOLORS['indianred4']
            mouse_over_colour = pygame.color.THECOLORS['indianred']
        if hint == UiHint.NO_DECORATION:
            return Button(rect, on_click, on_hover, label, disabled_label=disabled_label,
                          background_decoration=ButtonDecoration(background_colour),
                          mouse_over_decoration=ButtonDecoration(mouse_over_colour))
        else:
            return Button(rect, on_click, on_hover, label, disabled_label=disabled_label,
                          background_decoration=ButtonDecoration(background_colour),
                          mouse_over_decoration=ButtonDecoration(mouse_over_colour))

    def panel(self, rect, background_colour=None, layout: Optional[BaseLayout] = None, hint=UiHint.NORMAL):
        if background_colour is None:
            if hint == UiHint.WARNING:
                background_colour = pygame.color.THECOLORS['orange']
            elif hint == UiHint.ERROR:
                background_colour = pygame.color.THECOLORS['red']
        return Panel(rect, background_colour, decoration=BorderDecoration(self.colour), layout=layout)

    def menu(self, rect, background_colour=None, hint=UiHint.NORMAL):
        if background_colour is None:
            if hint == UiHint.WARNING:
                background_colour = pygame.color.THECOLORS['orange']
            elif hint == UiHint.ERROR:
                background_colour = pygame.color.THECOLORS['red']
        return Menu(rect, self, background_colour, decoration=BorderDecoration(self.colour))



from typing import Tuple, Callable

from pygame import Rect

from pyros_support_ui.components import CardsCollection, Button, BaseUIFactory, UiHint


class TextToggleButton(CardsCollection):
    def __init__(self, rect: Rect,
                 ui_factory: BaseUIFactory,
                 on_text: str, on_click_on: Callable[[Button, Tuple[int, int]], None],
                 off_text: str, on_click_off: Callable[[Button, Tuple[int, int]], None],
                 auto_toggle: bool = False,
                 hint: UiHint = UiHint.NORMAL):
        super().__init__(
            rect,
            [
                ("on", ui_factory.text_button(rect, on_text, on_click=self.callback_on, hint=hint)),
                ("off", ui_factory.text_button(rect, off_text, on_click=self.callback_off, hint=hint))
            ])
        self.on_click_on = on_click_on
        self.on_click_off = on_click_off
        self.auto_toggle = auto_toggle
        self.select_card("on")

    def callback_on(self, *_args) -> None:
        if self.auto_toggle:
            self.toggle_off()
        self.on_click_on(*_args)

    def callback_off(self, *_args) -> None:
        if self.auto_toggle:
            self.toggle_on()
        self.on_click_off(*_args)

    def is_toggle_on(self) -> bool:
        return self.selected_card_name == "on"

    def toggle_on(self) -> None:
        self.select_card("on")

    def toggle_off(self) -> None:
        self.select_card("off")

    def toggle(self, toggle: bool) -> None:
        if toggle:
            self.toggle_on()
        else:
            self.toggle_off()



from functools import partial
from typing import Optional, Callable, List, Tuple

import pygame
from pygame import Rect

from pyros_support_ui.components import BaseUIFactory, Collection


class DropDown(Collection):
    def __init__(self,
                 rect: Optional[Rect],
                 ui_factory: BaseUIFactory,
                 select_method: Optional[Callable] = None,
                 update_activation_button_text: Optional[bool] = True,
                 text: Optional[str] = None,
                 values: List[Tuple[str, str]] = None,
                 default_value: Optional[str] = None
                 ):
        super().__init__(rect)

        if text is None and (values is None or len(values) == 0 or len(values[0]) < 2):
            raise ValueError("One of 'text' or 'values' must be populated")

        if text is None:
            text = values[0][1]

        self.select_method = select_method
        self.update_activation_button_text = update_activation_button_text
        self._values = values if values is not None else []

        self.menu = ui_factory.menu(Rect(0, 0, 150, 200), background_colour=pygame.color.THECOLORS['black'])
        self.menu.hide()

        if values is not None:
            for ident, text in values:
                self.menu.add_menu_item(text, callback=partial(self.select_value, ident))

        self.drop_down_activation_button = ui_factory.text_button(self.rect, text, on_click=self.start_drop_down)
        self.add_component(self.drop_down_activation_button)
        if default_value is not None:
            self.select_value(default_value)

    def draw(self, surface):
        super().draw(surface)
        if self.menu.visible:
            self.menu.draw(surface)

    def add_item(self, item_id, item_text):
        self.menu.add_menu_item(item_text, callback=partial(self.select_value, item_id))
        self._values.append((item_id, item_text))

    def start_drop_down(self, button, _=None):
        if self.menu.rect.x != button.rect.x or self.menu.rect.y != button.rect.y:
            self.menu.redefine_rect(Rect(button.rect.x, button.rect.bottom, 200, 200))
        self.menu.show()

    def select_value(self, _id, _b=None, _p=None):
        self.menu.hide()
        if self.select_method is not None:
            self.select_method(_id)
        if self.update_activation_button_text:
            self.drop_down_activation_button.label.text = [v for v in self._values if v[0] == _id][0][1]



################################################################################
# Copyright (C) 2020 Abstract Horizon
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the Apache License v2.0
# which accompanies this distribution, and is available at
# https://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#    Daniel Sendula - initial API and implementation
#
#################################################################################
from enum import Enum, auto
from typing import Tuple, List, Optional

import pygame
from pygame import Rect
from pygame.font import Font


class ALIGNMENT(Enum):
    LEFT = auto()
    CENTER = auto()
    RIGHT = auto()
    TOP = auto()
    MIDDLE = auto()
    BOTTOM = auto()


class Component:
    def __init__(self, rect: Optional[Rect]):
        self.rect: Optional[Rect] = rect
        self.mouse_is_over = False
        self.font: Optional[Font] = None
        self._visible = True

    def _font(self) -> Optional[Font]:
        if self.font is None:
            self.font = pygame.font.SysFont('Arial', 30)
        return self.font

    @property
    def visible(self) -> bool: return self._visible

    @visible.setter
    def visible(self, visible: bool): self._visible = visible

    def draw(self, surface) -> None:
        pass

    def redefine_rect(self, rect: Rect) -> None:
        self.rect = rect

    def mouse_over(self, mouse_pos: Tuple[int, int]) -> None:
        self.mouse_is_over = True

    def mouse_left(self, mouse_pos: Tuple[int, int]) -> None:
        self.mouse_is_over = False

    def mouse_down(self, mouse_pos: Tuple[int, int]) -> None:
        pass

    def mouse_up(self, mouse_pos: Tuple[int, int]) -> None:
        pass

    def size(self) -> Tuple[int, int]:
        return self.rect.width, self.rect.height

    def key_down(self, code: str) -> bool:
        return False

    def key_up(self, code: str) -> bool:
        return False


class BaseLayout:
    def arrange(self, rect: Rect, components: List[Component]):
        for component in components:
            component.redefine_rect(rect)


class TopDownLayout(BaseLayout):
    def __init__(self, margin: int = 0):
        self.margin = margin

    def arrange(self, rect: Rect, components: List[Component]) -> None:
        y = rect.y
        for component in components:
            if component.rect is not None and component.rect.height != 0:
                height = component.rect.height
            else:
                height = (rect.height - self.margin * (len(components) - 1)) // len(components)
            component.redefine_rect(Rect(rect.x, y, rect.width, height))
            y += self.margin + height


class LeftRightLayout(BaseLayout):
    def __init__(self, margin: int = 0, h_alignment: ALIGNMENT = ALIGNMENT.LEFT):
        self.margin = margin
        self.h_alignment = h_alignment

    def arrange(self, rect: Rect, components: List[Component]) -> None:
        filled_in_space = False
        x = rect.x
        for component in components:
            if component.rect is not None and component.rect.width != 0:
                width = component.rect.width
            else:
                width = rect.width - self.margin * (len(components) - 1)
                filled_in_space = True
            component.redefine_rect(Rect(x, rect.y, width, rect.height))
            x += self.margin + width
        if not filled_in_space and self.h_alignment != ALIGNMENT.LEFT:
            offset = 0
            if self.h_alignment == ALIGNMENT.RIGHT:
                offset = rect.width - (x - rect.x) - 1
            elif self.h_alignment == ALIGNMENT.CENTER:
                offset = (rect.width - (x - rect.x) - 1) // 2
            for component in components:
                if component.rect is not None:
                    component.rect.x = component.rect.x + offset


class Collection(Component):
    def __init__(self, rect, layout: Optional[BaseLayout] = None, components: List[Component] = ()):
        super().__init__(rect)  # Call super constructor to store rectable
        self.components: List[Component] = []
        self.layout = layout
        self._selected_component: Optional[Component] = None
        for component in components:
            self.add_component(component)

    def add_component(self, component: Component) -> None:
        self.components.append(component)

    def remove_component(self, component: Component) -> None:
        i = self.components.index(component)
        if i > 0:
            del self.components[i]

    def draw(self, surface) -> None:
        for component in self.components:
            if component.visible:
                component.draw(surface)

    def find_component(self, pos) -> Optional[Component]:
        for component in reversed(self.components):
            rect = component.rect
            if component.visible and rect is not None and rect.collidepoint(pos):
                return component
        return None

    def redefine_rect(self, rect) -> None:
        self.rect = rect
        if self.layout is not None:
            self.layout.arrange(rect, self.components)
        else:
            for component in self.components:
                component.redefine_rect(rect)

    def mouse_over(self, mouse_pos) -> None:
        self.mouse_is_over = True
        component = self.find_component(mouse_pos)
        if component != self._selected_component and self._selected_component is not None:
            self._selected_component.mouse_left(mouse_pos)
        if component is not None:
            component.mouse_over(mouse_pos)
            self._selected_component = component

    def mouse_left(self, mouse_pos) -> None:
        self.mouse_is_over = False
        if self._selected_component is not None:
            self._selected_component.mouse_left(mouse_pos)

    def mouse_down(self, mouse_pos) -> None:
        component = self.find_component(mouse_pos)
        if component != self._selected_component and self._selected_component is not None:
            self._selected_component.mouse_left(mouse_pos)
        if component is not None:
            component.mouse_down(mouse_pos)
            self._selected_component = component

    def mouse_up(self, mouse_pos) -> None:
        if self._selected_component is not None:
            self._selected_component.mouse_up(mouse_pos)
            if not self._selected_component.rect.collidepoint(mouse_pos):
                # we released button outside of component - it would be nice to let it know mouse is not inside of it any more
                self._selected_component.mouse_left(mouse_pos)
        component = self.find_component(mouse_pos)
        if component is not None:
            # we released mouse over some other component - now it is turn for it to receive mouse over
            component.mouse_over(mouse_pos)


class Panel(Collection):
    def __init__(self, rect: Rect,
                 background_colour=None,
                 decoration: Optional[Component] = None,
                 layout: Optional[BaseLayout] = None,
                 horizontal_decoration_margin: int = 0,
                 vertical_decoration_margin: int = 0):
        super().__init__(rect, layout=layout)
        self.bacground_colour = background_colour
        self.decoration = decoration
        self.horizontal_decoration_margin = horizontal_decoration_margin
        self.vertical_decoration_margin = vertical_decoration_margin

    def redefine_rect(self, rect):
        super().redefine_rect(Rect(rect.left + self.horizontal_decoration_margin,
                                   rect.top + self.vertical_decoration_margin,
                                   rect.width - self.horizontal_decoration_margin * 2,
                                   rect.height - self.vertical_decoration_margin * 2))
        self.rect = rect
        if self.decoration is not None:
            self.decoration.redefine_rect(rect)

    def draw(self, surface):
        if self.bacground_colour is not None:
            pygame.draw.rect(surface, self.bacground_colour, self.rect)
        if self.decoration is not None:
            self.decoration.draw(surface)
        super().draw(surface)


class Menu(Panel):
    def __init__(self, rect, ui_factory, background_colour=None, decoration=None):
        super(Menu, self).__init__(ui_factory.ui_adapter.screen.get_rect())
        self.draw_rect = rect
        self.ui_factory = ui_factory
        self.bacground_colour = background_colour
        self.decoration = decoration
        self.menu_items = []
        self.visible = False
        self._key_selected = -1

    def redefine_rect(self, rect):
        self.draw_rect = rect
        self.rect = self.ui_factory.ui_adapter.screen.get_rect()
        self.calculate_height()
        if self.decoration is not None:
            self.decoration.redefine_rect(rect)

    def draw(self, surface):
        if self.bacground_colour is not None:
            pygame.draw.rect(surface, self.bacground_colour, self.draw_rect)
        if self._key_selected >= 0:
            for i in range(len(self.menu_items)):
                item = self.menu_items[i]

                item.mouse_is_over = i == self._key_selected

        super(Panel, self).draw(surface)
        if self.decoration is not None:
            self.decoration.draw(surface)

    def mouse_down(self, mouse_pos):
        if not self.draw_rect.collidepoint(mouse_pos):
            self.hide()
        else:
            super(Menu, self).mouse_over(mouse_pos)

    def show(self):
        self.visible = True
        self._clean_key_selected()

    def hide(self):
        self.visible = False
        self._clean_key_selected()

    def size(self):
        return self.draw_rect.width, self.draw_rect.height

    def _clean_key_selected(self):
        for item in self.menu_items:
            item.mouse_is_over = False

    def calculate_height(self):
        y = 5
        for item in self.menu_items:
            item.redefine_rect(Rect(self.draw_rect.x, y + self.draw_rect.y, self.draw_rect.width, item.rect.height))
            y += item.rect.height
        y += 5
        self.draw_rect.height = y

    def add_menu_item(self, label, callback=None, height=30):
        component = self.ui_factory.menu_item(Rect(0, 0, 0, height), label, callback)
        self.menu_items.append(component)
        self.add_component(component)
        self.calculate_height()

    def clear_menu_items(self):
        del self.menu_items[:]
        del self.components[:]
        self._key_selected = -1

    def key_down(self, code):
        if len(self.menu_items) > 0:
            if code == pygame.K_UP:
                if self._key_selected <= 0:
                    self._key_selected = len(self.menu_items) - 1
                else:
                    self._key_selected -= 1
                return True

            if code == pygame.K_DOWN:
                self._key_selected += 1
                if self._key_selected >= len(self.menu_items):
                    self._key_selected = 0
                return True

            if code == pygame.K_RETURN or code == pygame.K_KP_ENTER or code == pygame.K_SPACE:
                item = self.menu_items[self._key_selected]
                if isinstance(item, Button):
                    item.on_click(item, None)
                return True

        return False

    def key_up(self, code):
        return False


class CardsCollection(Collection):
    def __init__(self, rect, initial_cards: List[Tuple[str, Component]] = ()):
        super(CardsCollection, self).__init__(rect)
        self.cards = {}
        self.selected_card_name = None
        self.selected_card_component = None

        for name, component in initial_cards:
            self.add_card(name, component)

    def add_card(self, name, component) -> None:
        self.cards[name] = component
        component._visible = False
        super(CardsCollection, self).add_component(component)

    def card_component(self, name) -> Optional[Component]:
        return self.cards[name]

    def select_card(self, name) -> None:
        if name in self.cards:
            if self.selected_card_component is not None:
                self.selected_card_component.visible = False
            self.selected_card_component = self.cards[name]
            self.selected_card_component.visible = True
            self.selected_card_name = name
            return self.selected_card_component
        return None

    def selected_card_name(self):
        return self.selected_card_name

    def mouse_over(self, mouse_pos):
        self.selected_card_component.mouse_over(mouse_pos)

    def mouse_left(self, mouse_pos):
        self.selected_card_component.mouse_left(mouse_pos)

    def mouse_down(self, mouse_pos):
        self.selected_card_component.mouse_down(mouse_pos)

    def mouse_up(self, mouse_pos):
        self.selected_card_component.mouse_up(mouse_pos)


class Image(Component):
    def __init__(self, rect, surface, h_alignment=ALIGNMENT.LEFT, v_alignment=ALIGNMENT.TOP):
        super(Image, self).__init__(rect)  # Call super constructor to store rectable
        self._surface = surface
        self.h_alignment = h_alignment
        self.v_alignment = v_alignment

    def getImage(self):
        return self._surface

    def setImage(self, surface):
        self._surface = surface

    def draw(self, surface):
        if self._surface is not None:
            x = self.rect.x
            y = self.rect.y

            if self.h_alignment == ALIGNMENT.CENTER:
                x = self.rect.centerx - self._surface.get_width() // 2
            elif self.h_alignment == ALIGNMENT.RIGHT:
                x = self.rect.right - self._surface.get_width()

            if self.v_alignment == ALIGNMENT.MIDDLE:
                y = self.rect.centery - self._surface.get_height() // 2
            elif self.v_alignment == ALIGNMENT.BOTTOM:
                y = self.rect.bottom - self._surface.get_height()

            surface.blit(self._surface, (x, y))


class Label(Image):
    def __init__(self, rect, text, colour=None, font=None, h_alignment=ALIGNMENT.LEFT, v_alignment=ALIGNMENT.TOP):
        super(Label, self).__init__(rect, None, h_alignment, v_alignment)  # Call super constructor to store rectable
        self._text = text
        self.font = font
        self._colour = colour if colour is not None else pygame.color.THECOLORS['white']

    @property
    def text(self) -> str: return self._text

    @text.setter
    def text(self, text: str) -> None:
        if self._text != text:
            self._text = text
            self.invalidate_surface()

    @property
    def colour(self): return self._colour

    @colour.setter
    def colour(self, colour):
        self._colour = colour
        self.invalidate_surface()

    def invalidate_surface(self):
        self._surface = None

    def draw(self, surface):
        if self._surface is None:
            self._surface = self._font().render(self._text, False, self.colour)

        super(Label, self).draw(surface)


class Button(Component):
    def __init__(self,
                 rect,
                 on_click=None,
                 on_hover=None,
                 label=None,
                 disabled_label=None,
                 background_decoration=None,
                 mouse_over_decoration=None):
        super(Button, self).__init__(rect)  # Call super constructor to store rectable
        self.on_click = on_click
        self.on_hover = on_hover
        self._label = label
        self._disabled_label = disabled_label
        self.background_decoration = background_decoration
        self.mouse_over_decoration = mouse_over_decoration
        self.redefine_rect(rect)
        self._enabled = True

    @property
    def label(self) -> Component: return self._label

    @label.setter
    def label(self, label: Component):
        self._label = label
        self._label.redefine_rect(self.rect)

    @property
    def disabled_label(self) -> Component: return self._disabled_label

    @disabled_label.setter
    def disabled_label(self, label: Component):
        self._disabled_label = label
        self._disabled_label.redefine_rect(self.rect)

    def redefine_rect(self, rect):
        super(Button, self).redefine_rect(rect)
        if self._label is not None:
            self._label.redefine_rect(rect)  # set label's position to buttons
        if self._disabled_label is not None:
            self._disabled_label.redefine_rect(rect)  # set label's position to buttons
        if self.background_decoration is not None:
            self.background_decoration.redefine_rect(rect)
        if self.mouse_over_decoration is not None:
            self.mouse_over_decoration.redefine_rect(rect)

    def draw(self, surface):
        if self.mouse_is_over and self._enabled:
            if self.mouse_over_decoration is not None:
                self.mouse_over_decoration.draw(surface)
        else:
            if self.background_decoration is not None:
                self.background_decoration.draw(surface)

        if not self._enabled and self._disabled_label is not None:
            self._disabled_label.draw(surface)
        elif self._label is not None:  # this way 'label' can be anything - text, image or something custom drawn
            self._label.draw(surface)

    def mouse_up(self, mouse_pos):
        if self._enabled and self.rect.collidepoint(mouse_pos) and self.on_click is not None:
            self.on_click(self, mouse_pos)

    @property
    def enabled(self) -> bool: return self._enabled

    @enabled.setter
    def enabled(self, enabled: bool): self._enabled = enabled


class UIAdapter:
    def __init__(self, screen=None, screen_size=None, freq: int = 30, do_init: bool = True):
        self._screen = screen
        self.screen_size = None if screen_size is None else screen_size
        self._top_component = None
        self.mouse_is_down = False
        self.freq = freq
        self.frameclock = pygame.time.Clock()
        if self.screen is None and do_init:
            self.init()

    def init(self):
        pygame.init()
        self._screen = pygame.display.set_mode(self.screen_size)

    @property
    def screen(self): return self._screen

    @property
    def top_component(self) -> Component: return self._top_component

    @top_component.setter
    def top_component(self, top_component): self._top_component = top_component

    def process_event(self, event):
        if event.type == pygame.MOUSEMOTION:
            mouse_pos = pygame.mouse.get_pos()
            self.mouse_moved(mouse_pos)
        if event.type == pygame.MOUSEBUTTONDOWN and event.button == 1:
            mouse_pos = pygame.mouse.get_pos()
            self.mouse_down(mouse_pos)
        if event.type == pygame.MOUSEBUTTONUP and event.button == 1:
            mouse_pos = pygame.mouse.get_pos()
            self.mouse_up(mouse_pos)

    def mouse_moved(self, mouse_pos):
        if self.top_component is not None:
            self.top_component.mouse_over(mouse_pos)

    def mouse_down(self, mouse_pos):
        self.mouse_is_down = True
        if self.top_component is not None:
            self.top_component.mouse_down(mouse_pos)

    def mouse_up(self, mouse_pos):
        self.mouse_is_down = False
        if self.top_component is not None:
            self.top_component.mouse_up(mouse_pos)

    def draw(self, surface):
        if self.top_component is not None:
            self.top_component.draw(surface)

    def frame_end(self):
        pygame.display.flip()
        self.frameclock.tick(self.freq)


class UiHint:
    NO_DECORATION = auto()
    LIGHT = auto()
    NORMAL = auto()
    WARNING = auto()
    ERROR = auto()


class BorderDecoration(Component):
    def __init__(self, rect, colour):
        super(BorderDecoration, self).__init__(rect)  # Call super constructor to store rectable
        self.colour = colour

    def draw(self, surface):
        pygame.draw.rect(surface, self.colour, self.rect, 1)


class BaseUIFactory:
    def __init__(self, ui_adapter, font=None, small_font=None,
                 colour=None, warning_colour=None, error_colour=None,
                 disabled_colour=None, background_colour=None,
                 mouse_over_colour=None,
                 mouse_over_background_colour=None,
                 warning_mouse_over_background_colour=None,
                 error_mouse_over_background_colour=None):
        self._ui_adapter = ui_adapter
        self._font = font if font is not None else pygame.font.SysFont('Arial', 14)
        self._small_font = small_font if small_font is not None else pygame.font.SysFont('Arial', 9)
        self.colour = colour if colour is not None else pygame.color.THECOLORS['white']
        self.warning_colour = warning_colour if warning_colour else pygame.color.THECOLORS['orange']
        self.error_colour = error_colour if error_colour else pygame.color.THECOLORS['red']
        self.disabled_colour = disabled_colour if disabled_colour is not None else pygame.color.THECOLORS['gray']
        self.background_colour = background_colour if background_colour is not None else pygame.color.THECOLORS['black']
        self.mouse_over_colour = mouse_over_colour
        self.mouse_over_background_colour=mouse_over_background_colour if mouse_over_background_colour else pygame.color.THECOLORS['gray32']
        self.warning_mouse_over_background_colour=warning_mouse_over_background_colour if warning_mouse_over_background_colour else pygame.color.THECOLORS['darkorange4'],
        self.error_mouse_over_background_colour=error_mouse_over_background_colour if error_mouse_over_background_colour else pygame.color.THECOLORS['darkred']

    @property
    def ui_adapter(self) -> UIAdapter: return self._ui_adapter

    @property
    def font(self) -> Font: return self._font

    @property
    def small_font(self) -> Font: return self._small_font

    def label(self, rect, text, font=None, colour=None, h_alignment=ALIGNMENT.LEFT, v_alignment=ALIGNMENT.TOP, hint=UiHint.NORMAL) -> Label:
        # noinspection PyTypeChecker
        return None

    def _disabled_label(self, rect, text, font=None, h_alignment=ALIGNMENT.LEFT, v_alignment=ALIGNMENT.TOP, hint=UiHint.NORMAL) -> Label:
        # noinspection PyTypeChecker
        return None

    def image(self, rect, image, hint=UiHint.NORMAL) -> Image:
        # noinspection PyTypeChecker
        return None

    def button(self, rect, on_click=None, on_hover=None,
               label=None,
               disabled_label=None,
               hint=UiHint.NORMAL) -> Button:
        # noinspection PyTypeChecker
        return None

    def text_button(self, rect, text, on_click=None, on_hover=None, hint=UiHint.NORMAL, font=None) -> Button:
        return self.button(rect, on_click, on_hover,
                           self.label(None, text, h_alignment=ALIGNMENT.CENTER, v_alignment=ALIGNMENT.MIDDLE, font=font),
                           disabled_label=self._disabled_label(None, text, h_alignment=ALIGNMENT.CENTER, v_alignment=ALIGNMENT.MIDDLE, font=font),
                           hint=hint)

    def panel(self, rect, background_colour=None, layout: Optional[BaseLayout] = None, hint=UiHint.NORMAL) -> Panel:
        # noinspection PyTypeChecker
        return None

    def menu(self, rect, background_colour=None, hint=UiHint.NORMAL) -> Menu:
        # noinspection PyTypeChecker
        return None

    def _menu_item_text_button(self, rect, label, callback) -> Button:
        return self._menu_item_button(self, rect, self.label(None, label, h_alignment=ALIGNMENT.CENTER, v_alignment=ALIGNMENT.MIDDLE))

    def _menu_item_button(self, rect, label, callback) -> Button:
        return Button(rect, callback, label=label)

    def menu_item(self, rect, label, callback) -> Component:
        if isinstance(label, str):
            component = self._menu_item_text_button(rect, label, callback)
        elif isinstance(label, Label):
            component = self._menu_item_button(rect, label, callback)
        else:
            component = label

        return component

    def border(self, rect, colour=pygame.color.THECOLORS['white']) -> BorderDecoration:
        return BorderDecoration(rect, colour)




################################################################################
# Copyright (C) 2020 Abstract Horizon
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the Apache License v2.0
# which accompanies this distribution, and is available at
# https://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#    Daniel Sendula - initial API and implementation
#
#################################################################################

from pyros_support_ui.pyros_client_app import PyrosClientApp



################################################################################
# Copyright (C) 2020 Abstract Horizon
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the Apache License v2.0
# which accompanies this distribution, and is available at
# https://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#    Daniel Sendula - initial API and implementation
#
#################################################################################

import os
import time
import pyros

AGENT_PING_TIMEOUT = 10

_returncodes = {}

_agents = []

_lastPinged = 0

_agent_files = {}


def init(_client, filename, optional_files=None, agent_id=None):
    def send_file(_agent_id, _dest_path, _filename):
        with open(_filename, "rb") as f:
            _file_content = f.read()

            extra_name = os.path.join(_dest_path, os.path.split(_filename)[1])
            _agent_files[_agent_id].append(extra_name)

            pyros.publish("exec/" + _agent_id + "/process/" + extra_name, _file_content)

    def process_dir(_agent_id, _dest_path, _dir):
        for _file in os.listdir(_dir):
            if not _file.endswith('__pycache__'):
                if os.path.isdir(_file):
                    process_dir(_agent_id, os.path.join(_dest_path, _file), os.path.join(_dir, _file))
                else:
                    send_file(_agent_id, _dest_path, os.path.join(_dir, _file))

    # print("Connected to Rover " + str(client))
    if agent_id is None:
        if filename.endswith(".py"):
            agent_id = filename[: len(filename) - 3]
        else:
            agent_id = filename
            agent_id = agent_id.replace("/", "-")

        if "/" in agent_id:
            segments = agent_id.split("/")
            agent_id = segments[len(segments) - 1]

    with open(filename) as file:
        file_content = file.read()

    _agents.append(agent_id)
    _returncodes[agent_id] = None

    _agent_files[agent_id] = [agent_id + ".py"]

    pyros.subscribe("exec/" + str(agent_id) + "/out", process)
    pyros.subscribe("exec/" + str(agent_id) + "/status", process)
    pyros.publish("exec/" + str(agent_id), "stop")
    pyros.publish("exec/" + str(agent_id) + "/process", file_content)
    if optional_files is not None:
        for extra_file in optional_files:
            if os.path.isdir(extra_file):
                process_dir(agent_id, os.path.split(extra_file)[1], extra_file)
            else:
                send_file(agent_id, "", extra_file)

    pyros.publish("exec/" + str(agent_id), "make-agent")


def process(topic, message, _groups):
    if topic.startswith("exec/"):
        if topic.endswith("/out"):
            if len(message) > 0:
                agent_id = topic[5: len(topic) - 4]
                if message.endswith("\n"):
                    print(agent_id + ": " + message, end="")
                else:
                    print(agent_id + ": " + message)
            return True
        elif topic.endswith("/status"):
            agent_id = topic[5: len(topic) - 7]
            if message.startswith("stored "):
                filename = message[7:]
                if filename not in _agent_files[agent_id] and filename.endswith("_main.py"):
                    filename = filename[:-8] + ".py"
                if filename in _agent_files[agent_id]:
                    i = _agent_files[agent_id].index(filename)
                    del _agent_files[agent_id][i]
                    if len(_agent_files[agent_id]) == 0:
                        pyros.publish("exec/" + str(agent_id), "restart")
            elif message.startswith("exit"):
                if len(message) > 5:
                    _returncodes[agent_id] = message[5:]
                else:
                    _returncodes[agent_id] = "-1"

            return True

    return False


def stop_agent(_client, process_id):
    pyros.publish("exec/" + process_id, "stop")
    del _agents[process_id]


def keep_agents():
    global _lastPinged
    now = time.time()

    if now - _lastPinged > AGENT_PING_TIMEOUT:
        _lastPinged = now
        for agentId in _agents:
            pyros.publish("exec/" + agentId, "ping")


def returncode(agentId):
    return _returncodes[agentId]



################################################################################
# Copyright (C) 2020 Abstract Horizon
# All rights reserved. This program and the accompanying materials
# are made available under the terms of the Apache License v2.0
# which accompanies this distribution, and is available at
# https://www.apache.org/licenses/LICENSE-2.0
#
#  Contributors:
#    Daniel Sendula - initial API and implementation
#
#################################################################################
from typing import Union, List, Tuple, Optional, Callable

import pygame
import time

from pyros_support_ui.components import ALIGNMENT, Button, BaseUIFactory

NumberType = Union[int, float]
DataType = List[Tuple[float, NumberType]]


class GraphData:
    def __init__(self, maximum: NumberType, minimum: NumberType):
        self.maximum = maximum
        self.minimum = minimum
        self.name = ""

    def fetch(self, from_timestamp: float, to_timestamp: float) -> DataType:
        return []


class AutoScalingData(GraphData):
    def __init__(self, maximum: NumberType, minimum: NumberType):
        super(AutoScalingData, self).__init__(maximum, minimum)
        self._default_min = minimum
        self._default_max = maximum
        self._last_min_seen: Optional[NumberType] = None
        self._last_max_seen: Optional[NumberType] = None

    def auto_scale_data(self, data: DataType) -> None:
        max_seen = False
        max_found = -100000000
        min_seen = False
        min_found = 100000000
        for i in range(len(data)):
            if self.maximum < data[i][1]:
                self.maximum = data[i][1]
                self._last_max_seen = data[i][0]
                max_seen = True
            elif not max_seen and max_found < data[i][1]:
                max_found = data[i][1]
            if self.minimum > data[i][1]:
                self.minimum = data[i][1]
                self._last_min_seen = data[i][0]
                min_seen = True
            elif not min_seen and min_found > data[i][1]:
                min_found = data[i][1]
        if not max_seen:
            if max_found < self._default_max:
                max_found = self._default_max
            if max_found < self.maximum:
                self.maximum = max_found
        if not min_seen:
            if min_found > self._default_min:
                min_found = self._default_min
            if min_found > self.minimum:
                self.minimum = min_found


class TelemetryGraphData(AutoScalingData):
    def __init__(self,
                 telemetry_client,
                 stream,
                 column_name: str,
                 maximum: NumberType, minimum: NumberType, auto_scale: bool = False):
        super(TelemetryGraphData, self).__init__(maximum, minimum)
        self.telemetry_client = telemetry_client
        self.stream = stream
        self.name = column_name
        self.column_index = -1
        self.collect = True
        self.auto_scale = auto_scale

        for i, field in enumerate(self.stream.get_fields()):
            if field.name == column_name:
                self.column_index = i

    def fetch(self, from_timestamp: float, to_timestamp: float) -> DataType:
        result = []

        if self.collect:
            def fetch(data):
                result.extend([[d[0], d[self.column_index + 1]] for d in data])
                if self.auto_scale:
                    self.auto_scale_data(result)

            self.telemetry_client.retrieve(self.stream, from_timestamp, to_timestamp, fetch)
        return result


class ChangedSingleTelemetryGraphData(TelemetryGraphData):
    def __init__(self, telemetry_client, stream,
                 column_name: str, maximum: NumberType, minimum: NumberType,
                 transform: Callable[[NumberType], NumberType],
                 auto_scale: bool = False):
        super(ChangedSingleTelemetryGraphData, self).__init__(telemetry_client, stream, column_name, maximum, minimum, auto_scale=auto_scale)
        self.transform = transform

    def fetch(self, from_timestamp: float, to_timestamp: float) -> DataType:
        result = []

        if self.collect:
            def fetch(data):
                result.extend([[d[0], d[self.column_index + 1]] for d in data])
                for i in range(len(result)):
                    result[i][1] = self.transform(result[i][1])
                if self.auto_scale:
                    self.auto_scale_data(result)

            self.telemetry_client.retrieve(self.stream, from_timestamp, to_timestamp, fetch)
        return result


class GraphController:
    def __init__(self, timepoint: float = -1, timewindow: float = 10):
        self.timepoint = timepoint
        self.timewindow = timewindow

    def pause(self) -> None:
        self.timepoint = time.time() - self.timewindow

    def resume(self) -> None:
        self.timepoint = -1


class Graph(Button):
    # noinspection PyArgumentList
    def __init__(self, rect: pygame.Rect,
                 ui_factory: BaseUIFactory, small_font=None, controller: Optional[GraphController] = None,
                 on_click: Optional[Callable[[Tuple[int, int]], None]] = None,
                 inner_colour=pygame.color.THECOLORS['white'],
                 background_colour=pygame.color.THECOLORS['black'],
                 paused_background_colour=pygame.color.THECOLORS['black'],
                 label_colour=pygame.color.THECOLORS['grey']):
        self.label_colour = label_colour
        self.small_font = small_font if small_font is not None else ui_factory.small_font
        self.title = ui_factory.label(None, "", colour=self.label_colour, h_alignment=ALIGNMENT.CENTER, v_alignment=ALIGNMENT.TOP)
        self.max_label = ui_factory.label(None, "", colour=self.label_colour, h_alignment=ALIGNMENT.LEFT, v_alignment=ALIGNMENT.TOP)
        self.min_label = ui_factory.label(None, "", colour=self.label_colour, h_alignment=ALIGNMENT.LEFT, v_alignment=ALIGNMENT.BOTTOM)
        self.now_label = ui_factory.label(None, "now", colour=self.label_colour, h_alignment=ALIGNMENT.RIGHT, v_alignment=ALIGNMENT.BOTTOM)
        self.time_label = ui_factory.label(None, "", colour=self.label_colour, h_alignment=ALIGNMENT.CENTER, v_alignment=ALIGNMENT.BOTTOM)

        super(Graph, self).__init__(rect, on_click=self.button_cliecked)
        self.supplied_on_click = on_click
        self._graph_data: Optional[GraphData] = None
        self.border_colour = ui_factory.colour
        self.paused_border_colour = ui_factory.warning_colour
        self.inner_colour = inner_colour
        self.background_colour = background_colour
        self.paused_background_colour = paused_background_colour
        self.line_colour = (96, 96, 96)
        self.inner_rect = None
        self.units = ''
        self.controller = controller if controller is not None else self

        self.timepoint = -1
        self.timewindow = 10
        self._min_value = 0
        self._max_value = 0

    def button_cliecked(self, *_args):
        if self.supplied_on_click is not None:
            self.supplied_on_click(*_args)
        else:
            self.paused = not self.paused

    @property
    def paused(self):
        return self.controller.timepoint != -1

    @paused.setter
    def paused(self, paused: bool):
        if paused:
            self.controller.pause()
        else:
            self.controller.resume()

    def get_timepoint(self):
        return self.timepoint

    def get_timewindow(self):
        return self.timewindow

    def redefine_rect(self, rect):
        super(Graph, self).redefine_rect(rect)
        if rect is not None:
            self.inner_rect = rect.inflate(-3, -2)
            if self.title is not None:
                self.title.redefine_rect(self.inner_rect)
            if self.max_label is not None:
                self.max_label.redefine_rect(self.inner_rect)
            if self.min_label is not None:
                self.min_label.redefine_rect(self.inner_rect)
            if self.now_label is not None:
                self.now_label.redefine_rect(self.inner_rect)
            if self.time_label is not None:
                self.time_label.redefine_rect(self.inner_rect)

    @property
    def graph_data(self) -> GraphData:
        return self._graph_data

    @graph_data.setter
    def graph_data(self, graph_data: GraphData) -> None:
        self._graph_data = graph_data
        self.title.text = graph_data.name
        self._max_value = graph_data.maximum
        self._min_value = graph_data.minimum
        self._set_max_label(graph_data.maximum)
        self._set_min_label(graph_data.minimum)

    def _set_max_label(self, _max):
        self.max_label.text = f"{round(_max, 3)}"

    def _set_min_label(self, _min):
        self.min_label.text = f"{round(_min, 3)}"

    def draw(self, surface):
        if self.paused:
            pygame.draw.rect(surface, self.paused_border_colour, self.rect, 1)
        else:
            pygame.draw.rect(surface, self.border_colour, self.rect, 1)

        pygame.draw.rect(surface, self.background_colour, self.inner_rect)

        if self._graph_data is not None:

            timepoint = self.controller.timepoint
            timewindow = self.controller.timewindow

            timepoint = timepoint if timepoint >= 0 else time.time() - timewindow

            data = self._graph_data.fetch(timepoint, timepoint + timewindow + 0.01)
            graph_last_timepoint = timepoint + timewindow

            if self._max_value != self._graph_data.maximum:
                self._set_max_label(self._graph_data.maximum)
                self._max_value = self._graph_data.maximum
            if self._min_value != self._graph_data.minimum:
                self._set_min_label(self._graph_data.minimum)
                self._min_value = self._graph_data.minimum

            if len(data) > 0:
                t0 = data[0][0]
                now = time.time()

                t_minutes = now - t0
                if t_minutes < 0.1:
                    self.time_label.text = ""
                elif int(t_minutes) < 60:
                    self.time_label.text = str(int(t_minutes)) + " s"
                elif int(t_minutes) == 60:
                    self.time_label.text = "1 min"
                else:
                    self.time_label.text = str(int(t_minutes / 60)) + " mins"

                data_time_width = now - t0
                # if data_time_width < self.min_width_time:
                #     data_time_width = self.min_width_time
                # t_max = t0 + data_width
                # d_max = self.max_value

                if data_time_width <= 20:
                    minute_line_time = 1
                elif data_time_width <= 60:
                    minute_line_time = 5
                elif data_time_width <= 300:
                    minute_line_time = 25
                else:
                    minute_line_time = 300

                if data_time_width > timewindow:
                    data_time_width = timewindow

                tlast = data[-1][0]
                while t0 < tlast - data_time_width:
                    del data[0]
                    t0 = data[0][0]

                t = t0 + minute_line_time
                while t < tlast:
                    x = self.inner_rect.right - (graph_last_timepoint - t) * self.inner_rect.width / timewindow
                    pygame.draw.line(surface, self.line_colour, (x, self.inner_rect.y + 1), (x, self.inner_rect.bottom - 2), 1)
                    t += minute_line_time

                data_range = self._graph_data.maximum - self._graph_data.minimum

                if self._graph_data.minimum <= 0:
                    y = int(self.inner_rect.bottom + self._graph_data.minimum * self.inner_rect.height / data_range)
                    pygame.draw.line(surface, self.line_colour, (self.inner_rect.x, y), (self.inner_rect.right, y), 1)

                points = []
                for d in data:
                    t = d[0]
                    p = d[1]
                    # print(f"{t} : {p}")
                    if p > self._graph_data.maximum:
                        p = self._graph_data.maximum
                    elif p < self._graph_data.minimum:
                        p = self._graph_data.minimum

                    p -= self._graph_data.minimum

                    x = int(self.inner_rect.right - (graph_last_timepoint - t) * self.inner_rect.width / timewindow)
                    y = int(self.inner_rect.bottom - p * self.inner_rect.height / data_range)
                    points.append((x, y))

                if len(points) > 1:
                    pygame.draw.lines(surface, self.inner_colour, False, points)

                # points.append((x, self.inner_rect.bottom))
                # points.append((self.inner_rect.x, self.inner_rect.bottom))
                # pygame.draw.polygon(surface, self.inner_colour, points)
                # pygame.draw.polygon(surface, self.border_colour, points, 1)

                # if self.warning_value >= 0:
                #     y = self.inner_rect.bottom - self.warning_value * self.inner_rect.height / self.max_value
                #     pygame.draw.line(surface, self.warning_colour, (self.inner_rect.x + 1, y), (self.inner_rect.right - 2, y))
                #
                # if self.critical_value >= 0:
                #     y = self.inner_rect.bottom - self.critical_value * self.inner_rect.height / self.max_value
                #     pygame.draw.line(surface, self.critical_colour, (self.inner_rect.x + 1, y), (self.inner_rect.right - 2, y))

            self.title.draw(surface)
            self.max_label.draw(surface)
            self.min_label.draw(surface)
            self.now_label.draw(surface)
            self.time_label.draw(surface)
